/**
 * @file parens.h
 * prototypes for parens.cpp
 *
 * @author  Ben Gardner
 * @license GPL v2+
 */
#ifndef PARENS_H_INCLUDED
#define PARENS_H_INCLUDED

#include "uncrustify_types.h"


//! add parenthesis
void do_parens();


void do_parens_assign();


void do_parens_return();


#endif /* PARENS_H_INCLUDED */
