/**
 * @file detect.h
 * prototypes for detect.cpp
 *
 * @author  Ben Gardner
 * @license GPL v2+
 */
#ifndef DETECT_H_INCLUDED
#define DETECT_H_INCLUDED


//! Call all the detect_xxxx() functions
void detect_options();


#endif /* DETECT_H_INCLUDED */
