/**
 * @file mark_question_colon.h
 *
 * @author  Guy Maurel
 * @license GPL v2+
 */

#ifndef MARK_QUESTION_COLON_H_INCLUDED
#define MARK_QUESTION_COLON_H_INCLUDED

void mark_question_colon();

#endif /* MARK_QUESTION_COLON_H_INCLUDED */
