/**
 * @file align_left_shift.h
 *
 * @author  Guy Maurel
 * split from align.cpp
 * @author  Ben Gardner
 * @license GPL v2+
 */

#ifndef ALIGN_LEFT_SHIFT_H_INCLUDED
#define ALIGN_LEFT_SHIFT_H_INCLUDED

//! Align left shift operators '<<' (CT_SHIFT)
void align_left_shift();

#endif /* ALIGN_LEFT_SHIFT_H_INCLUDED */
