/**
 * @file align.h
 * prototypes for align.cpp
 *
 * @author  Ben Gardner
 * @license GPL v2+
 */
#ifndef ALIGN_H_INCLUDED
#define ALIGN_H_INCLUDED

#include "uncrustify_types.h"


void align_all();

#endif /* ALIGN_H_INCLUDED */
