/*
 * Generated by CMake
 */
#ifndef TOKEN_NAMES_H_INCLUDED
#define TOKEN_NAMES_H_INCLUDED

const char *token_names[] =
{
   "NONE",
   "PARENT_NOT_SET",
   "EOF",
   "UNKNOWN",
   "JUNK",
   "WHITESPACE",
   "SPACE",
   "NEWLINE",
   "NL_CONT",
   "FORM_FEED",
   "COMMENT_CPP",
   "COMMENT",
   "COMMENT_MULTI",
   "COMMENT_EMBED",
   "COMMENT_START",
   "COMMENT_END",
   "COMMENT_WHOLE",
   "COMMENT_CPP_ENDIF",
   "COMMENT_ENDIF",
   "IGNORED",
   "WORD",
   "NUMBER",
   "NUMBER_FP",
   "STRING",
   "STRING_MULTI",
   "IF",
   "ELSE",
   "ELSEIF",
   "CONSTEXPR",
   "FOR",
   "WHILE",
   "WHILE_OF_DO",
   "SWITCH",
   "CASE",
   "DO",
   "SYNCHRONIZED",
   "VOLATILE",
   "TYPEDEF",
   "STRUCT",
   "ENUM",
   "ENUM_CLASS",
   "SIZEOF",
   "DECLTYPE",
   "RETURN",
   "BREAK",
   "UNION",
   "GOTO",
   "CONTINUE",
   "C_CAST",
   "CPP_CAST",
   "D_CAST",
   "TYPE_CAST",
   "TYPENAME",
   "TEMPLATE",
   "PARAMETER_PACK",
   "WHERE_SPEC",
   "ASSIGN",
   "ASSIGN_NL",
   "SASSIGN",
   "ASSIGN_DEFAULT_ARG",
   "ASSIGN_FUNC_PROTO",
   "COMPARE",
   "SCOMPARE",
   "BOOL",
   "SBOOL",
   "ARITH",
   "SARITH",
   "SHIFT",
   "CARET",
   "DEREF",
   "INCDEC_BEFORE",
   "INCDEC_AFTER",
   "MEMBER",
   "DC_MEMBER",
   "C99_MEMBER",
   "INV",
   "DESTRUCTOR",
   "NOT",
   "D_TEMPLATE",
   "ADDR",
   "NEG",
   "POS",
   "STAR",
   "PLUS",
   "MINUS",
   "AMP",
   "BYREF",
   "POUND",
   "PREPROC",
   "PREPROC_INDENT",
   "PREPROC_BODY",
   "PP",
   "ELLIPSIS",
   "RANGE",
   "NULLCOND",
   "SEMICOLON",
   "VSEMICOLON",
   "COLON",
   "ASM_COLON",
   "CASE_COLON",
   "CASE_ELLIPSIS",
   "CLASS_COLON",
   "CONSTR_COLON",
   "D_ARRAY_COLON",
   "COND_COLON",
   "WHERE_COLON",
   "QUESTION",
   "COMMA",
   "ASM",
   "ATTRIBUTE",
   "AUTORELEASEPOOL",
   "OC_AVAILABLE",
   "OC_AVAILABLE_VALUE",
   "CATCH",
   "WHEN",
   "WHERE",
   "CLASS",
   "DELETE",
   "EXPORT",
   "FRIEND",
   "NAMESPACE",
   "PACKAGE",
   "NEW",
   "OPERATOR",
   "OPERATOR_VAL",
   "ASSIGN_OPERATOR",
   "ACCESS",
   "ACCESS_COLON",
   "THROW",
   "NOEXCEPT",
   "TRY",
   "BRACED_INIT_LIST",
   "USING",
   "USING_STMT",
   "USING_ALIAS",
   "D_WITH",
   "D_MODULE",
   "SUPER",
   "DELEGATE",
   "BODY",
   "DEBUG",
   "DEBUGGER",
   "INVARIANT",
   "UNITTEST",
   "UNSAFE",
   "FINALLY",
   "FIXED",
   "IMPORT",
   "D_SCOPE",
   "D_SCOPE_IF",
   "LAZY",
   "D_MACRO",
   "D_VERSION",
   "D_VERSION_IF",
   "PAREN_OPEN",
   "PAREN_CLOSE",
   "ANGLE_OPEN",
   "ANGLE_CLOSE",
   "SPAREN_OPEN",
   "SPAREN_CLOSE",
   "PPAREN_OPEN",
   "PPAREN_CLOSE",
   "FPAREN_OPEN",
   "FPAREN_CLOSE",
   "LPAREN_OPEN",
   "LPAREN_CLOSE",
   "TPAREN_OPEN",
   "TPAREN_CLOSE",
   "RPAREN_OPEN",
   "RPAREN_CLOSE",
   "BRACE_OPEN",
   "BRACE_CLOSE",
   "VBRACE_OPEN",
   "VBRACE_CLOSE",
   "SQUARE_OPEN",
   "SQUARE_CLOSE",
   "TSQUARE",
   "MACRO_OPEN",
   "MACRO_CLOSE",
   "MACRO_ELSE",
   "LABEL",
   "LABEL_COLON",
   "FUNCTION",
   "FUNC_CALL",
   "FUNC_CALL_USER",
   "FUNC_DEF",
   "FUNC_TYPE",
   "FUNC_VAR",
   "FUNC_PROTO",
   "FUNC_START",
   "FUNC_CLASS_DEF",
   "FUNC_CLASS_PROTO",
   "FUNC_CTOR_VAR",
   "FUNC_WRAP",
   "PROTO_WRAP",
   "MACRO_FUNC",
   "MACRO_FUNC_CALL",
   "MACRO",
   "QUALIFIER",
   "EXTERN",
   "DECLSPEC",
   "ALIGN",
   "TYPE",
   "PTR_TYPE",
   "TYPE_WRAP",
   "CPP_LAMBDA",
   "CPP_LAMBDA_RET",
   "EXECUTION_CONTEXT",
   "TRAILING_RET",
   "BIT_COLON",
   "ENUM_COLON",
   "OC_DYNAMIC",
   "OC_END",
   "OC_IMPL",
   "OC_INTF",
   "OC_PROTOCOL",
   "OC_PROTO_LIST",
   "OC_GENERIC_SPEC",
   "OC_PROPERTY",
   "OC_CLASS",
   "OC_CLASS_EXT",
   "OC_CATEGORY",
   "OC_SCOPE",
   "OC_MSG",
   "OC_MSG_CLASS",
   "OC_MSG_FUNC",
   "OC_MSG_NAME",
   "OC_MSG_SPEC",
   "OC_MSG_DECL",
   "OC_RTYPE",
   "OC_ATYPE",
   "OC_COLON",
   "OC_DICT_COLON",
   "OC_SEL",
   "OC_SEL_NAME",
   "OC_BLOCK",
   "OC_BLOCK_ARG",
   "OC_BLOCK_TYPE",
   "OC_BLOCK_EXPR",
   "OC_BLOCK_CARET",
   "OC_AT",
   "OC_PROPERTY_ATTR",
   "PP_DEFINE",
   "PP_DEFINED",
   "PP_INCLUDE",
   "PP_IF",
   "PP_ELSE",
   "PP_ENDIF",
   "PP_ASSERT",
   "PP_EMIT",
   "PP_ENDASM",
   "PP_ENDINPUT",
   "PP_ERROR",
   "PP_FILE",
   "PP_LINE",
   "PP_SECTION",
   "PP_ASM",
   "PP_UNDEF",
   "PP_PROPERTY",
   "PP_BODYCHUNK",
   "PP_PRAGMA",
   "PP_REGION",
   "PP_ENDREGION",
   "PP_REGION_INDENT",
   "PP_IF_INDENT",
   "PP_IGNORE",
   "PP_OTHER",
   "CHAR",
   "DEFINED",
   "FORWARD",
   "NATIVE",
   "STATE",
   "STOCK",
   "TAGOF",
   "DOT",
   "TAG",
   "TAG_COLON",
   "LOCK",
   "AS",
   "IN",
   "BRACED",
   "THIS",
   "BASE",
   "DEFAULT",
   "GETSET",
   "GETSET_EMPTY",
   "CONCAT",
   "CS_SQ_STMT",
   "CS_SQ_COLON",
   "CS_PROPERTY",
   "SQL_EXEC",
   "SQL_BEGIN",
   "SQL_END",
   "SQL_WORD",
   "SQL_ASSIGN",
   "CONSTRUCT",
   "LAMBDA",
   "ASSERT",
   "ANNOTATION",
   "FOR_COLON",
   "DOUBLE_BRACE",
   "CNG_HASINC",
   "CNG_HASINCN",
   "Q_EMIT",
   "Q_FOREACH",
   "Q_FOREVER",
   "Q_GADGET",
   "Q_OBJECT",
   "MODE",
   "DI",
   "HI",
   "QI",
   "SI",
   "NOTHROW",
   "WORD_",
};

#endif /* TOKEN_NAMES_H_INCLUDED */
